# SPDX-FileCopyrightText: 2019-2024 Greenbone AG
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

"""
Plugin API for handling git related tasks
"""

import os
import subprocess
from enum import Enum
from os import PathLike
from pathlib import Path
from tempfile import NamedTemporaryFile
from types import TracebackType
from typing import Any, Iterable, Iterator, List, Optional, Type

from autohooks.utils import GitError, exec_git, get_project_root_path

__all__ = [
    "exec_git",
    "GitError",
    "get_staged_status",
    "get_status",
    "is_partially_staged_status",
    "is_staged_status",
    "stage_files",
    "stash_unstaged_changes",
    "StatusEntry",
    "Status",
]


def _get_git_toplevel_path():
    try:
        git_dir = exec_git("rev-parse", "--show-toplevel").rstrip()
    except subprocess.CalledProcessError as e:
        print("could not determine toplevel directory. {e.output.decode()}")
        raise e from None
    return Path(git_dir).resolve()


class Status(Enum):
    """
    Status of a file in git
    """

    UNMODIFIED = " "
    MODIFIED = "M"
    ADDED = "A"
    DELETED = "D"
    RENAMED = "R"
    COPIED = "C"
    UPDATED = "U"
    UNTRACKED = "?"
    IGNORED = "!"


class StatusEntry:
    """
    Status of a file in the git index and working tree.

    Implements the :py:class:`os.PathLike` protocol.

    Attributes:
        index: Status in the index
        working_tree: Status in the working tree
        path: Path to the file
        root_path: An optional path to a root directory
        old_path: Set for renamed files
    """

    def __init__(
        self, status_string: str, root_path: Optional[Path] = None
    ) -> None:
        status = status_string[:2]
        filename = status_string[3:]

        # Status in the index
        self.index = Status(status[0])
        # Status in the working directory
        self.working_tree = Status(status[1])
        self.root_path = root_path

        if self.index == Status.RENAMED:
            new_filename, old_filename = filename.split("\0")
            self.path = Path(new_filename)
            self.old_path = Path(old_filename)
        else:
            # path of the file in git
            self.path = Path(filename)

    def __str__(self) -> str:
        return f"{self.index.value}{self.working_tree.value} {str(self.path)}"

    def __repr__(self) -> str:
        return f"<StatusEntry {str(self)}>"

    def absolute_path(self) -> Path:
        """
        Returns the absolute path of the file of this StatusEntry
        """
        if self.root_path:
            return (self.root_path / self.path).resolve()
        return self.path.resolve()

    def __fspath__(self):
        return self.path.__fspath__()


def _parse_status(output: str) -> Iterator[str]:
    output = output.rstrip("\0")
    if not output:
        return

    output_list = output.split("\0")
    while output_list:
        line = output_list.pop(0)
        if line[0] == Status.RENAMED.value:
            yield f"{line}\0{output_list.pop(0)}"
        else:
            yield line


def is_staged_status(status: StatusEntry) -> bool:
    """Returns true, if the status of the given :py:class:`StatusEntry` is
    staged.

    Arguments:
        status: A :py:class:`StatusEntry` object that contains the filename,
            path and the git status.

    Returns:
        True if file is staged, False else.
    """
    return (
        status.index != Status.UNMODIFIED
        and status.index != Status.UNTRACKED
        and status.index != Status.IGNORED
        and status.index != Status.DELETED
    )


def is_partially_staged_status(status: StatusEntry) -> bool:
    """Returns true, if the status of the given :py:class:`StatusEntry`
    is partially staged.

    Arguments:
        status: A :py:class:`StatusEntry` object that contains the filename,
            path and the git status.

    Returns:
        True if file is partially staged, False else.
    """
    return (
        status.index != Status.UNMODIFIED
        and status.index != Status.UNTRACKED
        and status.index != Status.IGNORED
        and status.index != Status.DELETED
        and status.working_tree != Status.UNMODIFIED
        and status.working_tree != Status.UNTRACKED
        and status.working_tree != Status.IGNORED
    )


def get_status(files: Optional[Iterable[PathLike]] = None) -> List[StatusEntry]:
    """Get information about the current git status.

    Arguments:
        files: (optional) specify an iterable of :py:class:`os.PathLike` and
            exclude all other paths for the status.

    Returns:
        A list of :py:class:`StatusEntry` instances that contain the status of
        the specific files.
    """
    args = [
        "status",
        "-z",
        "--ignore-submodules",
        "--untracked-files=no",
    ]

    if files is not None:
        args.append("--")
        args.extend([os.fspath(f) for f in files])

    output = exec_git(*args)
    root_path = _get_git_toplevel_path()
    return [StatusEntry(f, root_path) for f in _parse_status(output)]


def get_staged_status(
    files: Optional[Iterable[PathLike]] = None,
) -> List[StatusEntry]:
    """Get a list of :py:class:`StatusEntry` instances containing only staged
    files.

    Arguments:
        files: (optional) specify an iterable of files and exclude all other
            paths for the status.

    Returns:
        A list of :py:class:`StatusEntry` instances with files that are staged.
    """
    status = get_status(files)
    return [s for s in status if is_staged_status(s)]


def stage_files_from_status_list(status_list: Iterable[StatusEntry]) -> None:
    """Add the passed files from the status list to git staging index

    Deprecated. Please use stage_files instead.

    Arguments:
        status_list: A List of StatusEntry instances that should be added
    """
    stage_files(status_list)


def stage_files(files: Iterable[PathLike]) -> None:
    """Add the passed :py:class:`os.PathLike` to git staging index

    Arguments:
        files: An iterable of :py:class:`os.PathLike` to add to the index
    """
    filenames = [os.fspath(f) for f in files]
    exec_git("add", *filenames)


def get_diff(files: Optional[Iterable[StatusEntry]] = None) -> str:
    """Get the diff of the passed files

    Arguments:
        status_list: A List of StatusEntry instances that should be diffed

    Returns:
        string containing the diff of the given files
    """
    args = ["--no-pager", "diff"]

    if files is not None:
        args.append("--")
        args.extend([str(f.absolute_path()) for f in files])

    return exec_git(*args)


def _write_tree() -> str:
    """
    Create a tree object from the current index

    Returns:
        The name of the new tree object
    """
    return exec_git("write-tree").strip()


def _read_tree(ref_or_hashid: str) -> None:
    """
    Loads (reads) tree information into the index

    Arguments:
        ref_or_hashid: Git hash or ref to load into the index
    """
    exec_git("read-tree", ref_or_hashid)


def _checkout_from_index(files: Iterable[PathLike]) -> None:
    """
    Copy all files listed from the index to the working directory

    Arguments:
        files: Iterable of files that should be checked out into the working
               directory
    """
    filenames = [os.fspath(s) for s in files]
    exec_git("checkout-index", "-f", "--", *filenames)


def _set_ref(name: str, hashid: str) -> None:
    """
    Create a git ref for a hash

    Arguments:
        name: Name of the reference to be created
        hashid: Hash that the reference should point to
    """
    exec_git("update-ref", name, hashid)


def _get_tree_diff(tree1: str, tree2: str) -> bytes:
    """
    Calculate a diff between two tree objects

    Returns:
        The diff as bytes
    """
    return subprocess.check_output(
        [
            "git",
            "diff-tree",
            "--ignore-submodules",
            "--binary",
            "--no-color",
            "--no-ext-diff",
            "--unified=0",
            tree1,
            tree2,
        ]
    )


def _apply_diff(patch: bytes) -> None:
    with NamedTemporaryFile(mode="wb", buffering=0) as f:
        f.write(patch)

        exec_git(
            "apply",
            "-v",
            "--whitespace=nowarn",
            "--reject",
            "--recount",
            "--unidiff-zero",
            f.name,
        )


INDEX_REF = "refs/autohooks/index"
WORKING_REF = "refs/autohooks/working"


class stash_unstaged_changes:  # pylint: disable=invalid-name
    """
    A context manager that stashes changes that
        - are not staged, and
        - affect files that are partially staged.
    Changes that are made before the context manager exits, are added to the
    index.

    The stashed changes are restored when the context manager exits.

    Example: ::

        with stash_unstaged_changes():
            do_something()
    """

    def __init__(self, files: Optional[Iterable[PathLike]] = None) -> None:
        """
        Args:
            files: Optional iterable of path like objects to consider for being
                staged. By default all files in the git status are considered.
        """
        status_list = get_status(files)
        self.partially_staged = [
            s for s in status_list if is_partially_staged_status(s)
        ]

    def _stash_changes(self) -> None:
        # save current staging area aka. index
        self.index = _write_tree()
        # add ref to be able to restore index manually
        _set_ref(INDEX_REF, self.index)
        # add changes from files to index
        stage_files(self.partially_staged)
        # save index as working tree
        # unstaged changes are stored in the working tree now
        self.working_tree = _write_tree()
        # add ref to be able to restore working tree manually
        _set_ref(WORKING_REF, self.working_tree)

        # restore index without working tree changes
        # working tree changes are "stashed" now
        _read_tree(self.index)
        _checkout_from_index(self.partially_staged)

    def _restore_working_tree(self) -> None:
        # restore working tree
        _read_tree(self.working_tree)
        # checkout working tree
        _checkout_from_index(self.partially_staged)

    def __enter__(self) -> None:
        if self.partially_staged:
            self._stash_changes()

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> Any:
        if not self.partially_staged:
            return

        if exc_type is not None:
            # an error has occurred
            # restore working tree and index as it was before formatting
            self._restore_working_tree()
            _read_tree(self.index)
        else:
            # save possible changes made to the index
            changed_tree = _write_tree()

            self._restore_working_tree()

            # restore index
            _read_tree(changed_tree)

            # create and apply diff between index before running the plugin and
            # changes made by the plugin if some changes have been applied and
            # staged.
            # changed_tree will be the same as index if no changes are applied.
            # changed_tree may be the same as the working tree. in that case no
            # further action is needed.
            if changed_tree != self.index and changed_tree != self.working_tree:
                # create diff between working tree and changed tree
                patch = _get_tree_diff(self.index, changed_tree)
                try:
                    # apply diff to working tree
                    _apply_diff(patch)
                except GitError as e:
                    print(
                        "Found conflicts between plugin and local changes. "
                        "Plugin changes will be ignored for conflicted hunks.",
                        e,
                    )

                    rootpath = get_project_root_path()
                    for path in rootpath.glob("*.rej"):
                        path.unlink()



# SPDX-FileCopyrightText: 2022-2024 Greenbone AG
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

import unittest
from argparse import Namespace
from unittest.mock import MagicMock, call

from autohooks.cli.plugins import add_plugins, list_plugins, remove_plugins
from tests import temp_file, tempdir, unload_module


class AddPluginsCliTestCase(unittest.TestCase):
    def test_add_plugin(self):
        term = MagicMock()
        args = Namespace(name=["foo", "bar"])

        expected = """[tool.autohooks]
mode = "pythonpath"
pre-commit = ["bar", "foo"]
"""

        with tempdir(change_into=True) as tmp_dir:
            add_plugins(term, args)

            term.warning.assert_not_called()
            term.info.assert_has_calls([call("Added plugins:")])
            term.ok.assert_has_calls([call('"bar"'), call('"foo"')])

            pyproject_toml = tmp_dir / "pyproject.toml"
            content = pyproject_toml.read_text(encoding="utf8")

            self.assertEqual(content, expected)

    def test_add_plugin_with_existing(self):
        term = MagicMock()
        args = Namespace(name=["foo"])
        existing = """[tool.autohooks]
mode = "poetry"
pre-commit = ["bar"]
"""

        expected = """[tool.autohooks]
mode = "poetry"
pre-commit = ["bar", "foo"]
"""

        with temp_file(
            existing, name="pyproject.toml", change_into=True
        ) as pyproject_toml:
            add_plugins(term, args)

            term.warning.assert_not_called()
            term.info.assert_has_calls(
                [call("Added plugins:"), call("Currently used plugins:")]
            )
            term.ok.assert_has_calls([call('"foo"')])

            content = pyproject_toml.read_text(encoding="utf8")

            self.assertEqual(content, expected)

    def test_add_duplicate_plugin(self):
        term = MagicMock()
        args = Namespace(name=["foo", "bar"])
        existing = """[tool.autohooks]
mode = "poetry"
pre-commit = ["bar"]
"""

        expected = """[tool.autohooks]
mode = "poetry"
pre-commit = ["bar", "foo"]
"""

        with temp_file(
            existing, name="pyproject.toml", change_into=True
        ) as pyproject_toml:
            add_plugins(term, args)

            term.warning.assert_has_calls([call('"bar"')])
            term.info.assert_has_calls(
                [
                    call("Skipped already used plugins:"),
                    call("Added plugins:"),
                    call("Currently used plugins:"),
                ]
            )
            term.ok.assert_has_calls([call('"foo"')])

            content = pyproject_toml.read_text(encoding="utf8")

            self.assertEqual(content, expected)


class ListPluginsCliTestCase(unittest.TestCase):
    def setUp(self) -> None:
        unload_module("bar")
        return super().setUp()

    def test_invalid_plugins(self):
        term = MagicMock()
        args = Namespace()

        existing = """[tool.autohooks]
mode = "poetry"
pre-commit = ["bar"]
"""
        with temp_file(existing, name="pyproject.toml", change_into=True):
            list_plugins(term, args)

            term.warning.assert_not_called()
            term.info.assert_has_calls(
                [
                    call("Currently used plugins:"),
                ]
            )
            term.ok.assert_not_called()
            term.error.assert_called_once_with(
                '"bar": "bar" is not a valid autohooks plugin. No module '
                "named 'bar'"
            )

    def test_existing_plugins(self):
        term = MagicMock()
        args = Namespace()

        pyproject_toml_content = """[tool.autohooks]
mode = "poetry"
pre-commit = ["bar"]
"""

        plugin_content = """
def precommit(config=None, **kwargs):
    pass
"""
        with tempdir(change_into=True, add_to_sys_path=True) as tmp_dir:
            pyproject_toml = tmp_dir / "pyproject.toml"
            pyproject_toml.write_text(pyproject_toml_content, encoding="utf8")

            bar_plugin = tmp_dir / "bar.py"
            bar_plugin.write_text(plugin_content, encoding="utf8")

            list_plugins(term, args)

            term.warning.assert_not_called()
            term.info.assert_has_calls(
                [
                    call("Currently used plugins:"),
                ]
            )
            term.error.assert_not_called()
            term.ok.assert_called_once_with('"bar"')


class RemovePluginsCliTestCase(unittest.TestCase):
    def test_remove_plugins(self):
        term = MagicMock()
        args = Namespace(name=["foo", "bar"])

        existing = """[tool.autohooks]
mode = "pythonpath"
pre-commit = ["bar", "foo"]
"""
        expected = """[tool.autohooks]
mode = "pythonpath"
pre-commit = []
"""

        with temp_file(
            existing, name="pyproject.toml", change_into=True
        ) as pyproject_toml:
            remove_plugins(term, args)

            term.warning.assert_not_called()
            term.info.assert_has_calls([call("Removed plugins:")])
            term.ok.assert_has_calls([call('"bar"'), call('"foo"')])

            content = pyproject_toml.read_text(encoding="utf8")

            self.assertEqual(content, expected)

    def test_no_config(self):
        term = MagicMock()
        args = Namespace(name=["foo", "bar"])

        with tempdir(change_into=True) as temp_dir:
            remove_plugins(term, args)

            term.warning.assert_called_once_with("No plugins to remove.")
            term.info.assert_not_called()
            term.ok.assert_not_called()

            pyproject_toml = temp_dir / "pyproject.toml"
            self.assertFalse(pyproject_toml.exists())

    def test_remove_plugins_skipped(self):
        term = MagicMock()
        args = Namespace(name=["foo", "bar"])

        existing = """[tool.autohooks]
mode = "pythonpath"
pre-commit = ["foo"]
"""
        expected = """[tool.autohooks]
mode = "pythonpath"
pre-commit = []
"""

        with temp_file(
            existing, name="pyproject.toml", change_into=True
        ) as pyproject_toml:
            remove_plugins(term, args)

            term.warning.assert_called_once_with('"bar"')
            term.info.assert_has_calls(
                [
                    call("Skipped not used plugins:"),
                    call("Removed plugins:"),
                    call("Currently used plugins:"),
                ]
            )
            term.ok.assert_has_calls([call('"foo"')])

            content = pyproject_toml.read_text(encoding="utf8")

            self.assertEqual(content, expected)



# SPDX-FileCopyrightText: 2022-2024 Greenbone AG
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

import os
import sys
import tempfile
from contextlib import AbstractContextManager, contextmanager
from pathlib import Path
from typing import Generator

from autohooks.utils import exec_git


class AddSysPath(AbstractContextManager):
    """
    Context Manager to add a directory path to the module search path aka.
    sys.path. The directory path

    Example:
        with AddSysPath("")
    """

    def __init__(self, directory: Path):
        self.directory = str(directory.resolve())

        if sys.path[0] != self.directory:
            sys.path.insert(0, self.directory)

    def cleanup(self):
        try:
            sys.path.remove(self.directory)
        except ValueError:
            # self.directory was not in the path
            pass

    def __exit__(self, __exc_type, __exc_value, __traceback) -> None:
        self.cleanup()


@contextmanager
def tempdir(
    change_into: bool = False, add_to_sys_path: bool = False
) -> Generator[Path, None, None]:
    """
    Context Manager to create a temporary directory

    Args:
        change_into: Set the created temporary as the current working directory
        add_to_sys_path: Add the created temporary directory to the directories
            for searching for Python modules

    Returns:
        A path to the created temporary directory
    """
    temp_dir = tempfile.TemporaryDirectory()
    dir_path = Path(temp_dir.name)

    if change_into:
        os.chdir(dir_path)

    if add_to_sys_path:
        sys_path = AddSysPath(dir_path)

    yield Path(dir_path)

    if add_to_sys_path:
        sys_path.cleanup()

    temp_dir.cleanup()


@contextmanager
def tempgitdir() -> Generator[Path, None, None]:
    temp_dir = tempfile.TemporaryDirectory()
    temp_path = Path(temp_dir.name)

    os.chdir(str(temp_path))

    exec_git("init", "-b", "main")
    exec_git("config", "--local", "user.email", "max.mustermann@example.com")
    exec_git("config", "--local", "user.name", "Max Mustermann")

    yield temp_path

    temp_dir.cleanup()


@contextmanager
def temp_file(
    content: str,
    *,
    name: str = "test.toml",
    change_into: bool = False,
) -> Generator[Path, None, None]:
    """
    A Context Manager to create a temporary file within a new temporary
    directory. The temporary file and directory are removed when the context is
    exited.

    Args:
        content: Content to write into the temporary file.
        name: Name of the temporary file. "test.toml" by default.
        change_into: Adjust the current working directory to the temporary
            directory.
    """
    with tempdir(change_into=change_into) as tmp_dir:
        test_file = tmp_dir / name
        test_file.write_text(content, encoding="utf8")
        yield test_file


@contextmanager
def temp_python_module(
    content: str, *, name="foo"
) -> Generator[Path, None, None]:
    """
    A Context Manager to create a new Python module in a temporary directory.
    The temporary directory will be added to the module search path and removed
    from the search path when the context is exited. Also it is ensured that
    the module is unloaded if the context is exited.

    Args:
        content: Python code to write into the temporary module.
        name: Name of the new Python module. By default: "foo".

    Example:
        with temp_python_module("print()", name="bar") as python_module_path
    """
    with (
        tempdir(
            add_to_sys_path=True,
        ) as tmp_dir,
        ensure_unload_module(name),
    ):
        test_file = tmp_dir / f"{name}.py"
        test_file.write_text(content, encoding="utf8")
        yield test_file


def unload_module(name: str) -> None:
    """
    Unload a Python module
    """
    if name in sys.modules:
        del sys.modules[name]


@contextmanager
def ensure_unload_module(name: str) -> Generator[None, None, None]:
    """
    Ensure that a module gets removed even if an error occurs

    Example:
        with ensure_unload_module("foo.bar"):
            do_something()
    """
    yield
    unload_module(name)



# SPDX-FileCopyrightText: 2019-2024 Greenbone AG
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

from argparse import Namespace
from pathlib import Path

from autohooks.config import (
    AUTOHOOKS_SECTION,
    get_pyproject_toml_path,
    load_config_from_pyproject_toml,
)
from autohooks.hooks import PreCommitHook
from autohooks.precommit.run import (
    CheckPluginError,
    CheckPluginWarning,
    autohooks_module_path,
    check_plugin,
)
from autohooks.settings import Mode
from autohooks.terminal import Terminal


# pylint: disable=unused-argument
def check_hooks(term: Terminal, args: Namespace) -> None:
    pre_commit_hook = PreCommitHook()
    check_pre_commit_hook(term, pre_commit_hook)

    pyproject_toml = get_pyproject_toml_path()

    check_config(term, pyproject_toml, pre_commit_hook)


def check_pre_commit_hook(
    term: Terminal, pre_commit_hook: PreCommitHook
) -> None:
    if pre_commit_hook.exists():
        if pre_commit_hook.is_autohooks_pre_commit_hook():
            term.ok("autohooks pre-commit hook is active.")

            if pre_commit_hook.is_current_autohooks_pre_commit_hook():
                term.ok("autohooks pre-commit hook is up-to-date.")
            else:
                term.warning(
                    "autohooks pre-commit hook is outdated. Please run "
                    "'autohooks activate --force' to update your pre-commit "
                    "hook."
                )

            hook_mode = pre_commit_hook.read_mode()
            if hook_mode == Mode.UNKNOWN:
                term.warning(
                    f"Unknown autohooks mode in {str(pre_commit_hook)}. "
                    f'Falling back to "{str(hook_mode.get_effective_mode())}" '
                    "mode."
                )
        else:
            term.error(
                "autohooks pre-commit hook is not active. But a different "
                f"pre-commit hook has been found at {str(pre_commit_hook)}."
            )

    else:
        term.error(
            "autohooks pre-commit hook not active. Please run 'autohooks "
            "activate'."
        )


def check_config(
    term: Terminal,
    pyproject_toml: Path,
    pre_commit_hook: PreCommitHook,
) -> None:
    if not pyproject_toml.exists():
        term.error(
            f"Missing {str(pyproject_toml)} file. Please add a pyproject.toml "
            f'file and include a "{AUTOHOOKS_SECTION}" section.'
        )
    else:
        config = load_config_from_pyproject_toml(pyproject_toml)
        if not config.has_autohooks_config():
            term.error(
                f"autohooks is not enabled in your {str(pyproject_toml)} file."
                f' Please add a "{AUTOHOOKS_SECTION}" section.'
            )
        elif pre_commit_hook.exists():
            config_mode = config.get_mode()
            hook_mode = pre_commit_hook.read_mode()

            if config_mode == Mode.UNDEFINED:
                term.warning(
                    f"autohooks mode is not defined in {str(pyproject_toml)}."
                )
            elif config_mode == Mode.UNKNOWN:
                term.warning(
                    f"Unknown autohooks mode in {str(pyproject_toml)}."
                )

            elif (
                config_mode.get_effective_mode()
                != hook_mode.get_effective_mode()
            ):
                term.warning(
                    f'autohooks mode "{str(hook_mode)}" in pre-commit '
                    f"hook {str(pre_commit_hook)} differs from "
                    f'mode "{str(config_mode)}" in {str(pyproject_toml)}.'
                )

            term.info(
                f'Using autohooks mode "{str(hook_mode.get_effective_mode())}".'
            )

            plugins = config.get_pre_commit_script_names()
            if not plugins:
                term.error(
                    "No autohooks plugin is activated in "
                    f"{str(pyproject_toml)} for your pre commit hook. Please "
                    'add a "pre-commit = [plugin1, plugin2]" setting.'
                )
            else:
                with autohooks_module_path():
                    for name in plugins:
                        result = check_plugin(name)
                        if result:
                            if isinstance(result, CheckPluginError):
                                term.error(str(result))
                            elif isinstance(result, CheckPluginWarning):
                                term.warning(str(result))
                            else:
                                term.info(str(result))
                        else:
                            term.ok(f'Plugin "{name}" active and loadable.')



# SPDX-FileCopyrightText: 2019-2024 Greenbone AG
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

import unittest
from pathlib import Path

from autohooks.config import (
    AutohooksConfig,
    Config,
    Mode,
    load_config_from_pyproject_toml,
)


def get_test_config_path(name):
    return Path(__file__).parent / name


class AutohooksConfigTestCase(unittest.TestCase):
    def test_load_from_toml_file(self):
        config_path = get_test_config_path("pyproject.test1.toml")
        self.assertTrue(config_path.is_file())

        config = load_config_from_pyproject_toml(config_path)

        self.assertTrue(config.has_autohooks_config())

        self.assertListEqual(
            config.get_pre_commit_script_names(), ["foo", "bar"]
        )

    def test_load_from_non_existing_toml_file(self):
        config_path = Path("foo")
        self.assertFalse(config_path.exists())

        config = load_config_from_pyproject_toml(config_path)

        self.assertFalse(config.has_autohooks_config())
        self.assertEqual(config.get_mode(), Mode.UNDEFINED)

        self.assertEqual(len(config.get_pre_commit_script_names()), 0)

    def test_load_from_string(self):
        config = AutohooksConfig.from_string(
            """
            [tool.autohooks]
            pre-commit = ["foo", "bar"]
            """
        )

        self.assertTrue(config.has_autohooks_config())

        self.assertListEqual(
            config.get_pre_commit_script_names(), ["foo", "bar"]
        )

    def test_load_from_empty_string(self):
        config = AutohooksConfig.from_string("")

        self.assertFalse(config.has_autohooks_config())

        self.assertEqual(config.get_pre_commit_script_names(), [])

    def test_empty_config(self):
        config = AutohooksConfig()

        self.assertFalse(config.has_autohooks_config())
        self.assertEqual(config.get_mode(), Mode.UNDEFINED)

        self.assertEqual(len(config.get_pre_commit_script_names()), 0)

    def test_empty_config_dict(self):
        config = AutohooksConfig.from_dict({"foo": "bar"})

        self.assertFalse(config.has_autohooks_config())
        self.assertEqual(config.get_mode(), Mode.UNDEFINED)

        self.assertEqual(len(config.get_pre_commit_script_names()), 0)

    def test_missing_pre_commit(self):
        config = AutohooksConfig.from_dict(
            {"tool": {"autohooks": {"foo": "bar"}}}
        )

        self.assertTrue(config.has_autohooks_config())
        self.assertEqual(config.get_mode(), Mode.UNDEFINED)

        self.assertEqual(len(config.get_pre_commit_script_names()), 0)

    def test_get_mode_pipenv(self):
        config = AutohooksConfig.from_dict(
            {"tool": {"autohooks": {"mode": "pipenv"}}}
        )

        self.assertTrue(config.has_autohooks_config())
        self.assertEqual(config.get_mode(), Mode.PIPENV)

        self.assertEqual(len(config.get_pre_commit_script_names()), 0)

    def test_get_mode_pipenv_multiline(self):
        config = AutohooksConfig.from_dict(
            {"tool": {"autohooks": {"mode": "pipenv_multiline"}}}
        )

        self.assertTrue(config.has_autohooks_config())
        self.assertEqual(config.get_mode(), Mode.PIPENV_MULTILINE)

        self.assertEqual(len(config.get_pre_commit_script_names()), 0)

    def test_get_mode_poetry(self):
        config = AutohooksConfig.from_dict(
            {"tool": {"autohooks": {"mode": "poetry"}}}
        )

        self.assertTrue(config.has_autohooks_config())
        self.assertEqual(config.get_mode(), Mode.POETRY)

        self.assertEqual(len(config.get_pre_commit_script_names()), 0)

    def test_get_mode_poetry_multiline(self):
        config = AutohooksConfig.from_dict(
            {"tool": {"autohooks": {"mode": "poetry_multiline"}}}
        )

        self.assertTrue(config.has_autohooks_config())
        self.assertEqual(config.get_mode(), Mode.POETRY_MULTILINE)

        self.assertEqual(len(config.get_pre_commit_script_names()), 0)

    def test_get_mode_pythonpath(self):
        config = AutohooksConfig.from_dict(
            {"tool": {"autohooks": {"mode": "pythonpath"}}}
        )

        self.assertTrue(config.has_autohooks_config())
        self.assertEqual(config.get_mode(), Mode.PYTHONPATH)

        self.assertEqual(len(config.get_pre_commit_script_names()), 0)

    def test_get_mode_unknown(self):
        config = AutohooksConfig.from_dict(
            {"tool": {"autohooks": {"mode": "foo"}}}
        )

        self.assertTrue(config.has_autohooks_config())
        self.assertEqual(config.get_mode(), Mode.UNKNOWN)

        self.assertEqual(len(config.get_pre_commit_script_names()), 0)

    def test_get_mode_undefined(self):
        config = AutohooksConfig.from_dict(
            {"tool": {"autohooks": {"mode": None}}}
        )

        self.assertTrue(config.has_autohooks_config())
        self.assertEqual(config.get_mode(), Mode.UNDEFINED)

        self.assertEqual(len(config.get_pre_commit_script_names()), 0)

    def test_get_config_dict(self):
        config_in = {"tool": {"autohooks": {"lorem": "ipsum"}}, "foo": "bar"}
        config = AutohooksConfig.from_dict(config_in)

        self.assertEqual(config.get_mode(), Mode.UNDEFINED)

        config_out = config.get_config()

        self.assertEqual(config_out.get_value("foo"), "bar")

    def test_load_config_dict_from_toml_file(self):
        config_path = get_test_config_path("pyproject.test2.toml")
        self.assertTrue(config_path.is_file())

        autohooksconfig = load_config_from_pyproject_toml(config_path)
        self.assertTrue(autohooksconfig.has_autohooks_config())

        config = autohooksconfig.get_config()

        self.assertEqual(
            config.get("tool")
            .get("autohooks")
            .get("plugins")
            .get("foo")
            .get_value("bar"),
            "ipsum",
        )
        self.assertEqual(
            config.get("tool")
            .get("autohooks")
            .get("plugins")
            .get("foo")
            .get_value("lorem"),
            "dolor",
        )
        self.assertEqual(
            config.get("tool")
            .get("autohooks")
            .get("plugins")
            .get("bar")
            .get_value("foo"),
            "bar",
        )


class ConfigTestCase(unittest.TestCase):
    def test_empty_config(self):
        config = Config()

        self.assertTrue(config.is_empty())

        bar_config = config.get("foo").get("bar")
        self.assertTrue(bar_config.is_empty())

        self.assertTrue(config.get_value("foo", True))
        self.assertListEqual(config.get_value("foo", ["a"]), ["a"])

    def test_config_dict(self):
        config_dict = {
            "tool": {"autohooks": {"lorem": "ipsum"}},
            "foo": {"lorem": "ipsum"},
        }
        config = Config(config_dict)

        self.assertFalse(config.is_empty())

        foo_config = config.get("foo")

        self.assertFalse(foo_config.is_empty())

        self.assertEqual(foo_config.get_value("lorem"), "ipsum")
        self.assertEqual(foo_config.get_value("lorem", "dolor"), "ipsum")
        self.assertEqual(foo_config.get_value("bar", "dolor"), "dolor")

    def test_config_point_syntax(self):
        config_dict = {
            "tool": {"autohooks": {"plugins": {"bar": {"lorem": "ipsum"}}}}
        }

        config = Config(config_dict)

        self.assertFalse(config.is_empty())

        bar_config = config.get("tool", "autohooks", "plugins", "bar")
        self.assertFalse(bar_config.is_empty())
        self.assertEqual(bar_config.get_value("lorem"), "ipsum")

    def test_has_key(self):
        config = Config()
        self.assertFalse(config.has_key("foo"))
        config = Config({"foo": "bar"})
        self.assertTrue(config.has_key("foo"))


if __name__ == "__main__":
    unittest.main()



# SPDX-FileCopyrightText: 2019-2024 Greenbone AG
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

import os
import subprocess
import unittest
from pathlib import Path
from tempfile import TemporaryDirectory
from unittest.mock import MagicMock, patch

from autohooks.utils import (
    GitError,
    exec_git,
    get_git_directory_path,
    get_git_hook_directory_path,
    get_project_autohooks_plugins_path,
    get_project_root_path,
    get_pyproject_toml_path,
    is_project_root,
    is_split_env,
)
from tests import tempdir


class ExecGitTestCase(unittest.TestCase):
    def test_exec_fail(self):
        with (
            tempdir(change_into=True),
            self.assertRaisesRegex(
                GitError,
                r"Git command '\['git', 'foo'\]' returned non-zero exit "
                "status 1",
            ) as err,
        ):
            exec_git("foo")

        exception = err.exception
        self.assertEqual(exception.cmd, ["git", "foo"])
        self.assertEqual(exception.returncode, 1)

    def test_exec_ignore_error(self):
        with tempdir(change_into=True):
            self.assertEqual(exec_git("foo", ignore_errors=True), "")

    def test_exec_success(self):
        with tempdir(change_into=True):
            exec_git("init")


class GitHookDirPathTestCase(unittest.TestCase):
    def test_get_git_hook_directory_path(self):
        path = Path("foo")
        git_hook_dir_path = get_git_hook_directory_path(path)
        self.assertEqual(git_hook_dir_path, path / "hooks")

    def test_with_env_pwd(self):
        with TemporaryDirectory() as f:
            temp_path = Path(f)

            exec_git("-C", str(temp_path), "init")

            os.chdir(str(temp_path))

            git_dir_path = (temp_path / ".git").resolve()

            git_hook_dir_path = get_git_hook_directory_path()
            self.assertEqual(git_hook_dir_path, git_dir_path / "hooks")


class IsProjectRootTestCase(unittest.TestCase):
    def setUp(self):
        self.tempdir = TemporaryDirectory()
        self.temp_path = Path(self.tempdir.name)

    def tearDown(self):
        self.tempdir.cleanup()

    def test_no_root(self):
        self.assertFalse(is_project_root(self.temp_path))

    def test_pyproject_toml(self):
        pyproject_toml = self.temp_path / "pyproject.toml"
        pyproject_toml.touch()

        self.assertTrue(pyproject_toml.exists())
        self.assertTrue(is_project_root(self.temp_path))

    def test_dot_git(self):
        git_dir = self.temp_path / ".git"
        git_dir.mkdir()

        self.assertTrue(git_dir.exists())
        self.assertTrue(git_dir.is_dir())
        self.assertTrue(is_project_root(self.temp_path))

    def test_setup_py(self):
        setup_py = self.temp_path / "setup.py"
        setup_py.touch()

        self.assertTrue(setup_py.exists())
        self.assertTrue(is_project_root(self.temp_path))

    def test_setup_cfg(self):
        setup_cfg = self.temp_path / "setup.cfg"
        setup_cfg.touch()

        self.assertTrue(setup_cfg.exists())
        self.assertTrue(is_project_root(self.temp_path))


class GetProjectRootPath(unittest.TestCase):
    def setUp(self):
        self.tempdir = TemporaryDirectory()
        self.temp_path = Path(self.tempdir.name).resolve()

    def tearDown(self):
        self.tempdir.cleanup()

    def test_direct_root(self):
        setup_py = self.temp_path / "setup.py"
        setup_py.touch()

        root_path = get_project_root_path(self.temp_path)

        self.assertEqual(root_path, self.temp_path)

    def test_with_sub_directory(self):
        sub_path = self.temp_path / "foo"
        sub_path.mkdir()
        setup_py = self.temp_path / "setup.py"
        setup_py.touch()

        root_path = get_project_root_path(sub_path)

        self.assertEqual(root_path, self.temp_path)

    def test_without_project_root(self):
        root_path = get_project_root_path(self.temp_path)

        self.assertEqual(root_path, self.temp_path)

    def test_with_env_pwd(self):
        setup_py = self.temp_path / "setup.py"
        setup_py.touch()
        sub_path = self.temp_path / "foo"
        sub_path.mkdir()

        os.chdir(str(self.temp_path))

        root_path = get_project_root_path()

        self.assertEqual(root_path, self.temp_path)


class GetProjectAutohooksPluginsPathTestCase(unittest.TestCase):
    def setUp(self):
        self.tempdir = TemporaryDirectory()
        self.temp_path = Path(self.tempdir.name).resolve()

        setup_py = self.temp_path / "setup.py"
        setup_py.touch()

    def tearDown(self):
        self.tempdir.cleanup()

    def test_with_subpath(self):
        sub_path = self.temp_path / "foo"
        sub_path.mkdir()

        autohooks_plugins_path = get_project_autohooks_plugins_path(sub_path)
        self.assertEqual(autohooks_plugins_path, self.temp_path / ".autohooks")

    def test_with_env_pwd(self):
        os.chdir(str(self.temp_path))

        autohooks_plugins_path = get_project_autohooks_plugins_path()
        self.assertEqual(autohooks_plugins_path, self.temp_path / ".autohooks")


class GetPyProjectTomlPathTestCase(unittest.TestCase):
    def setUp(self):
        self.tempdir = TemporaryDirectory()
        self.temp_path = Path(self.tempdir.name).resolve()

    def tearDown(self):
        self.tempdir.cleanup()

    def test_with_pyproject_toml_file(self):
        pyproject_toml_file = self.temp_path / "pyproject.toml"
        pyproject_toml_file.touch()

        sub_path = self.temp_path / "foo"
        sub_path.mkdir()

        pyproject_toml_path = get_pyproject_toml_path(sub_path)
        self.assertEqual(pyproject_toml_path, self.temp_path / "pyproject.toml")

    def test_with_env_pwd(self):
        os.chdir(str(self.temp_path))

        setup_py = self.temp_path / "setup.py"
        setup_py.touch()

        pyproject_toml_path = get_pyproject_toml_path()
        self.assertEqual(pyproject_toml_path, self.temp_path / "pyproject.toml")


class GetGitDirectoryPath(unittest.TestCase):
    def setUp(self):
        self.tempdir = TemporaryDirectory()
        self.temp_path = Path(self.tempdir.name).resolve()

        exec_git("-C", str(self.temp_path), "init")

        self.git_dir_path = (self.temp_path / ".git").resolve()

        self.assertTrue(self.git_dir_path.exists())

    def tearDown(self):
        self.tempdir.cleanup()

    def test_with_root_dir(self):
        os.chdir(str(self.temp_path))

        git_dir_path = get_git_directory_path()
        self.assertEqual(git_dir_path, self.git_dir_path.resolve())

    def test_with_subdir(self):
        sub_path = self.temp_path / "foo"
        sub_path.mkdir()

        os.chdir(str(self.temp_path))

        git_dir_path = get_git_directory_path()
        self.assertEqual(git_dir_path, self.git_dir_path)

    def test_no_git_directory(self):
        with tempdir(change_into=True), self.assertRaises(GitError):
            get_git_directory_path()


class IsSplitEnvTestCase(unittest.TestCase):
    @patch("subprocess.run")
    def test_is_split_env(self, subprocess_mock: MagicMock):
        self.assertTrue(is_split_env())

        subprocess_mock.assert_called_once_with(
            ["/usr/bin/env", "-S", "echo", "True"],
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            universal_newlines=True,
            check=True,
        )

    @patch("subprocess.run")
    def test_is_not_split_env(self, subprocess_mock: MagicMock):
        subprocess_mock.side_effect = subprocess.CalledProcessError(1, ["foo"])
        self.assertFalse(is_split_env())

        subprocess_mock.assert_called_once_with(
            ["/usr/bin/env", "-S", "echo", "True"],
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            universal_newlines=True,
            check=True,
        )


if __name__ == "__main__":
    unittest.main()



# SPDX-FileCopyrightText: 2022-2024 Greenbone AG
#
# SPDX-License-Identifier: GPL-3.0-or-later
#


import unittest
from argparse import Namespace
from unittest.mock import MagicMock, call

from autohooks.cli.activate import install_hooks
from tests import tempgitdir

CONFIG = """
[tool.autohooks]
mode = "poetry"
"""


class ActivateCliTestCase(unittest.TestCase):
    def test_install_hooks(self):
        with tempgitdir() as tmpdir:
            pyproject_toml = tmpdir / "pyproject.toml"
            pyproject_toml.write_text(CONFIG, encoding="utf8")

            term = MagicMock()
            args = Namespace(force=False, mode=None)

            install_hooks(term, args)

            term.warning.assert_not_called()
            term.ok.assert_called_once_with(
                f"autohooks pre-commit hook installed at {tmpdir}/"
                ".git/hooks/pre-commit using poetry mode."
            )

    def test_install_exists(self):
        with tempgitdir() as tmpdir:
            pyproject_toml = tmpdir / "pyproject.toml"
            pyproject_toml.write_text(CONFIG, encoding="utf8")
            pre_commit = tmpdir / ".git" / "hooks" / "pre-commit"
            pre_commit.touch()

            term = MagicMock()
            args = Namespace(force=False, mode=None)

            install_hooks(term, args)

            term.ok.assert_not_called()
            term.warning.assert_called_once_with(
                f"autohooks pre-commit hook is already installed at {tmpdir}/"
                ".git/hooks/pre-commit."
            )
            term.info.assert_has_calls(
                (
                    call(
                        "Run 'autohooks activate --force' to override the "
                        "current installed pre-commit hook."
                    ),
                    call(
                        "Run 'autohooks check' to validate the current status "
                        "of the installed pre-commit hook."
                    ),
                )
            )

    def test_install_hooks_force(self):
        with tempgitdir() as tmpdir:
            pyproject_toml = tmpdir / "pyproject.toml"
            pyproject_toml.write_text(CONFIG, encoding="utf8")
            pre_commit = tmpdir / ".git" / "hooks" / "pre-commit"
            pre_commit.touch()

            term = MagicMock()
            args = Namespace(force=True, mode=None)

            install_hooks(term, args)

            term.warning.assert_not_called()
            term.ok.assert_has_calls(
                (
                    call(f"autohooks settings written to {pyproject_toml}."),
                    call(
                        f"autohooks pre-commit hook installed at {pre_commit}"
                        " using poetry mode."
                    ),
                )
            )

    def test_install_no_config(self):
        with tempgitdir() as tmpdir:
            term = MagicMock()
            args = Namespace(force=False, mode=None)

            install_hooks(term, args)

            term.warning.assert_not_called()
            term.ok.assert_has_calls(
                (
                    call(
                        "autohooks settings written to "
                        f"{tmpdir}/pyproject.toml."
                    ),
                    call(
                        f"autohooks pre-commit hook installed at {tmpdir}/"
                        ".git/hooks/pre-commit using pythonpath mode."
                    ),
                )
            )

    def test_install_hooks_with_mode(self):
        with tempgitdir() as tmpdir:
            pyproject_toml = tmpdir / "pyproject.toml"
            pyproject_toml.write_text(CONFIG, encoding="utf8")

            term = MagicMock()
            args = Namespace(force=False, mode="pythonpath")

            install_hooks(term, args)

            term.warning.assert_not_called()
            term.ok.assert_called_once_with(
                f"autohooks pre-commit hook installed at {tmpdir}/"
                ".git/hooks/pre-commit using pythonpath mode."
            )



# SPDX-FileCopyrightText: 2019-2024 Greenbone AG
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

import importlib
import inspect
import sys
from contextlib import contextmanager
from types import ModuleType
from typing import Generator, Optional

from rich.progress import TaskID

from autohooks.config import load_config_from_pyproject_toml
from autohooks.hooks import PreCommitHook
from autohooks.settings import Mode
from autohooks.terminal import Progress, Terminal, _set_terminal
from autohooks.utils import get_project_autohooks_plugins_path


@contextmanager
def autohooks_module_path() -> Generator:
    plugins = get_project_autohooks_plugins_path()
    plugins_dir_name = str(plugins)

    if plugins.is_dir():
        sys.path.append(plugins_dir_name)

    yield

    if plugins_dir_name in sys.path:
        sys.path.remove(plugins_dir_name)


def load_plugin(name: str) -> ModuleType:
    return importlib.import_module(name)


def has_precommit_function(plugin: ModuleType) -> bool:
    return hasattr(plugin, "precommit") and inspect.isfunction(plugin.precommit)


def has_precommit_parameters(plugin: ModuleType) -> bool:
    signature = inspect.signature(plugin.precommit)
    return bool(signature.parameters)


def check_hook_is_current(
    term: Terminal, pre_commit_hook: PreCommitHook
) -> None:
    if not pre_commit_hook.is_current_autohooks_pre_commit_hook():
        term.warning(
            "autohooks pre-commit hook is outdated. Please run "
            "'autohooks activate --force' to update your pre-commit "
            "hook."
        )


def check_hook_mode(term: Terminal, config_mode: Mode, hook_mode: Mode) -> None:
    if config_mode.get_effective_mode() != hook_mode.get_effective_mode():
        term.warning(
            f'autohooks mode "{str(hook_mode)}" in pre-commit hook differs '
            f'from mode "{str(config_mode)}" in pyproject.toml file.'
        )


class CheckPluginResult:
    def __init__(self, message: str) -> None:
        self.message = message

    def __str__(self) -> str:
        return self.message


class CheckPluginError(CheckPluginResult):
    """
    Raised if a plugin check failed
    """


class CheckPluginWarning(CheckPluginResult):
    """
    Used if a plugin check raises a warning
    """


def check_plugin(plugin_name: str) -> Optional[CheckPluginResult]:
    """
    Check if a plugin (Python module) is valid and can be used

    Returns:
        A CheckPluginResult in case of an issue with the plugin
    """
    try:
        plugin = load_plugin(plugin_name)
        if not has_precommit_function(plugin):
            return CheckPluginError(
                f'Plugin "{plugin_name}" has no precommit '
                "function. The function is required to run"
                " the plugin as git pre commit hook."
            )
        elif not has_precommit_parameters(plugin):
            return CheckPluginWarning(
                f'Plugin "{plugin_name}" uses a deprecated '
                "signature for its precommit function. It "
                "is missing the **kwargs parameter."
            )
    except ImportError as e:
        return CheckPluginError(
            f'"{plugin_name}" is not a valid autohooks plugin. {e}'
        )
    return None


class ReportProgress:
    """
    A class to report progress of a plugin
    """

    def __init__(self, progress: Progress, task_id: int) -> None:
        self._progress = progress
        self._task_id = TaskID(task_id)

    def init(self, total: int) -> None:
        """
        Init the progress with the total number to process

        Args:
            total: Most of the time this should be the number of files to
                process.
        """
        self._progress.update(self._task_id, total=total)

    def update(self, advance: int = 1) -> None:
        """
        Update the number of already processed steps/items/files.

        This increases the progress indicator.

        Args:
            advance: Number of steps/items/files the progress advanced. By
                default 1.
        """
        self._progress.advance(self._task_id, advance)


def run() -> int:
    term = Terminal()

    _set_terminal(term)

    config = load_config_from_pyproject_toml()

    pre_commit_hook = PreCommitHook()

    check_hook_is_current(term, pre_commit_hook)

    if config.has_autohooks_config():
        check_hook_mode(term, config.get_mode(), pre_commit_hook.read_mode())

    plugins = get_project_autohooks_plugins_path()
    plugins_dir_name = str(plugins)

    if plugins.is_dir():
        sys.path.append(plugins_dir_name)

    term.bold_info("autohooks => pre-commit")

    with (
        autohooks_module_path(),
        term.indent(),
        Progress(terminal=term) as progress,
    ):
        for name in config.get_pre_commit_script_names():
            term.info(f"Running {name}")
            with term.indent():
                try:
                    plugin = load_plugin(name)
                    if not has_precommit_function(plugin):
                        term.fail(
                            f"No precommit function found in plugin {name}. "
                            "Your autohooks settings may be invalid."
                        )
                        return 1

                    task_id = progress.add_task(
                        f"Running {name}", total=None, name=name
                    )
                    report_progress = ReportProgress(progress, task_id)
                    if has_precommit_parameters(plugin):
                        retval = plugin.precommit(
                            config=config.get_config(),
                            report_progress=report_progress,
                        )
                    else:
                        term.warning(
                            "precommit function without kwargs is deprecated. "
                            f"Please update {name} to a newer version."
                        )
                        retval = plugin.precommit()

                    progress.update(task_id, total=1, advance=1)

                    if retval:
                        return retval

                except ImportError as e:
                    term.error(
                        "An error occurred while importing pre-commit "
                        f"hook {name}. {e}."
                    )
                    return 1
                except Exception as e:  # pylint: disable=broad-except
                    term.error(
                        "An error occurred while running pre-commit "
                        f"hook {name}. {e}."
                    )
                    return 1

    return 0



# SPDX-FileCopyrightText: 2019-2024 Greenbone AG
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

# pylint: disable=invalid-name

import unittest

from autohooks.settings import AutohooksSettings, Mode
from tests import temp_file, tempdir


class ModeTestCase(unittest.TestCase):
    def test_get_effective_mode(self):
        self.assertEqual(Mode.PIPENV.get_effective_mode(), Mode.PIPENV)
        self.assertEqual(Mode.PYTHONPATH.get_effective_mode(), Mode.PYTHONPATH)
        self.assertEqual(
            Mode.PIPENV_MULTILINE.get_effective_mode(), Mode.PIPENV_MULTILINE
        )
        self.assertEqual(Mode.POETRY.get_effective_mode(), Mode.POETRY)
        self.assertEqual(
            Mode.POETRY_MULTILINE.get_effective_mode(), Mode.POETRY_MULTILINE
        )
        self.assertEqual(Mode.UNDEFINED.get_effective_mode(), Mode.PYTHONPATH)
        self.assertEqual(Mode.UNKNOWN.get_effective_mode(), Mode.PYTHONPATH)

    def test_get_pipenv_mode_from_string(self):
        self.assertEqual(Mode.from_string("pipenv"), Mode.PIPENV)
        self.assertEqual(Mode.from_string("PIPENV"), Mode.PIPENV)

    def test_get_pipenv_multiline_mode_from_string(self):
        self.assertEqual(
            Mode.from_string("pipenv_multiline"), Mode.PIPENV_MULTILINE
        )
        self.assertEqual(
            Mode.from_string("PIPENV_MULTILINE"), Mode.PIPENV_MULTILINE
        )

    def test_get_pythonpath_mode_from_string(self):
        self.assertEqual(Mode.from_string("pythonpath"), Mode.PYTHONPATH)
        self.assertEqual(Mode.from_string("PYTHONPATH"), Mode.PYTHONPATH)

    def test_get_poetry_mode_from_string(self):
        self.assertEqual(Mode.from_string("poetry"), Mode.POETRY)
        self.assertEqual(Mode.from_string("POETRY"), Mode.POETRY)

    def test_get_poetry_multiline_mode_from_string(self):
        self.assertEqual(
            Mode.from_string("poetry_multiline"), Mode.POETRY_MULTILINE
        )
        self.assertEqual(
            Mode.from_string("POETRY_MULTILINE"), Mode.POETRY_MULTILINE
        )

    def test_get_invalid_mode_from_string(self):
        self.assertEqual(Mode.from_string("foo"), Mode.UNKNOWN)
        self.assertEqual(Mode.from_string(None), Mode.UNDEFINED)
        self.assertEqual(Mode.from_string(""), Mode.UNDEFINED)

    def test_string_conversion(self):
        self.assertEqual(str(Mode.PIPENV), "pipenv")
        self.assertEqual(str(Mode.PIPENV_MULTILINE), "pipenv_multiline")
        self.assertEqual(str(Mode.PYTHONPATH), "pythonpath")
        self.assertEqual(str(Mode.POETRY_MULTILINE), "poetry_multiline")
        self.assertEqual(str(Mode.POETRY), "poetry")
        self.assertEqual(str(Mode.UNKNOWN), "unknown")
        self.assertEqual(str(Mode.UNDEFINED), "undefined")


pyproject_toml_1 = """
[tool.autohooks]
pre-commit = ['foo', 'bar']
plugins.foo.bar = 'ipsum'
plugins.foo.lorem = 'dolor'

[tool.autohooks.plugins.bar]
foo = 'bar'
"""


class AutohooksSettingsTestCase(unittest.TestCase):
    def test_create_empty(self):
        settings = AutohooksSettings()
        self.assertEqual(settings.mode, Mode.UNDEFINED)
        self.assertEqual(settings.pre_commit, [])

    def test_create(self):
        settings = AutohooksSettings(mode=Mode.POETRY, pre_commit=["a", "b"])

        self.assertEqual(settings.mode, Mode.POETRY)
        self.assertEqual(settings.pre_commit, ["a", "b"])

    def test_save_settings_new_empty(self):
        settings = AutohooksSettings()
        expected = """[tool.autohooks]
mode = "pythonpath"
pre-commit = []
"""
        with tempdir() as tmp_path:
            toml = tmp_path / "test.toml"
            settings.write(toml)

            self.assertEqual(settings.mode, Mode.UNDEFINED)
            self.assertEqual(settings.pre_commit, [])

            self.assertEqual(toml.read_text(encoding="utf8"), expected)

    def test_save_empty_override(self):
        settings = AutohooksSettings()
        expected = """
[tool.autohooks]
pre-commit = []
plugins.foo.bar = 'ipsum'
plugins.foo.lorem = 'dolor'
mode = "pythonpath"

[tool.autohooks.plugins.bar]
foo = 'bar'
"""
        with temp_file(pyproject_toml_1) as toml:
            settings.write(toml)

            self.assertEqual(settings.mode, Mode.UNDEFINED)
            self.assertEqual(settings.pre_commit, [])

            self.assertEqual(toml.read_text(encoding="utf8"), expected)

    def test_save_override(self):
        settings = AutohooksSettings(mode=Mode.POETRY, pre_commit=["a", "b"])
        expected = """
[tool.autohooks]
pre-commit = ["a", "b"]
plugins.foo.bar = 'ipsum'
plugins.foo.lorem = 'dolor'
mode = "poetry"

[tool.autohooks.plugins.bar]
foo = 'bar'
"""
        with temp_file(pyproject_toml_1) as toml:
            settings.write(toml)

            self.assertEqual(settings.mode, Mode.POETRY)
            self.assertEqual(settings.pre_commit, ["a", "b"])

            self.assertEqual(toml.read_text(encoding="utf8"), expected)


if __name__ == "__main__":
    unittest.main()



# SPDX-FileCopyrightText: 2022-2024 Greenbone AG
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

import sys
import unittest
from argparse import Namespace
from unittest.mock import MagicMock, call

from autohooks.cli.check import check_config, check_hooks, check_pre_commit_hook
from autohooks.config import AUTOHOOKS_SECTION
from autohooks.hooks import PreCommitHook, get_pre_commit_hook_path
from autohooks.settings import Mode
from autohooks.template import POETRY_SHEBANG, TEMPLATE_VERSION
from autohooks.utils import get_pyproject_toml_path
from tests import tempgitdir


class CheckCliTestCase(unittest.TestCase):
    def test_no_pre_commit_hooks_no_pyproject_toml(self):
        term = MagicMock()
        args = Namespace()

        with tempgitdir() as tmpdir:
            check_hooks(term, args)

        term.ok.assert_not_called()
        term.warning.assert_not_called()
        term.info.assert_not_called()
        self.assertEqual(term.error.call_count, 2)
        term.error.assert_has_calls(
            (
                call(
                    "autohooks pre-commit hook not active. Please run "
                    "'autohooks activate'."
                ),
                call(
                    f"Missing {tmpdir}/pyproject.toml file. Please add a "
                    'pyproject.toml file and include a "tool.autohooks" '
                    "section."
                ),
            )
        )

    def test_no_pre_commit_hooks_no_autohooks_settings(self):
        term = MagicMock()
        args = Namespace()

        with tempgitdir() as tmpdir:
            pyproject_toml = tmpdir / "pyproject.toml"
            pyproject_toml.touch()

            check_hooks(term, args)

        term.ok.assert_not_called()
        term.warning.assert_not_called()
        term.info.assert_not_called()
        self.assertEqual(term.error.call_count, 2)
        term.error.assert_has_calls(
            (
                call(
                    "autohooks pre-commit hook not active. Please run "
                    "'autohooks activate'."
                ),
                call(
                    f"autohooks is not enabled in your {tmpdir}/pyproject.toml "
                    'file. Please add a "tool.autohooks" section.'
                ),
            )
        )

    def test_all_checks_success(self):
        term = MagicMock()
        args = Namespace()

        with tempgitdir() as tmpdir:
            pyproject_toml = tmpdir / "pyproject.toml"
            pyproject_toml.write_text(
                """[tool.autohooks]
mode = "poetry"
pre-commit = ["plugin1"]
""",
                encoding="utf8",
            )
            pre_commit_hook = PreCommitHook()
            pre_commit_hook.write(mode=Mode.POETRY)
            dot_autohooks_dir = tmpdir / ".autohooks"
            dot_autohooks_dir.mkdir()
            plugin1 = dot_autohooks_dir / "plugin1.py"
            plugin1.write_text(
                """
def precommit(*args):
    pass
            """,
                encoding="utf8",
            )

            check_hooks(term, args)

        self.assertEqual(term.ok.call_count, 3)
        term.ok.assert_has_calls(
            (
                call("autohooks pre-commit hook is active."),
                call("autohooks pre-commit hook is up-to-date."),
                call('Plugin "plugin1" active and loadable.'),
            )
        )
        term.warning.assert_not_called()
        term.info.assert_called_once_with('Using autohooks mode "poetry".')
        term.error.assert_not_called()

        del sys.modules["plugin1"]


class CheckPreCommitHookTestCase(unittest.TestCase):
    def test_no_precommit_hook(self):
        term = MagicMock()

        with tempgitdir():
            pre_commit_hook = PreCommitHook()

            check_pre_commit_hook(term, pre_commit_hook)

        term.ok.assert_not_called()
        term.warning.assert_not_called()
        term.info.assert_not_called()
        term.error.assert_called_once_with(
            "autohooks pre-commit hook not active. Please run "
            "'autohooks activate'."
        )

    def test_no_autohooks_precommit_hook(self):
        term = MagicMock()

        with tempgitdir():
            pre_commit_hook_path = get_pre_commit_hook_path()
            pre_commit_hook_path.write_text(
                '#!/bin/sh\necho "Hello World\n"', encoding="utf8"
            )
            pre_commit_hook = PreCommitHook()

            check_pre_commit_hook(term, pre_commit_hook)

        term.ok.assert_not_called()
        term.warning.assert_not_called()
        term.info.assert_not_called()
        term.error.assert_called_once_with(
            "autohooks pre-commit hook is not active. But a different "
            f"pre-commit hook has been found at {pre_commit_hook_path}."
        )

    def test_outdated_precommit_hook(self):
        term = MagicMock()

        with tempgitdir():
            pre_commit_hook_path = get_pre_commit_hook_path()
            pre_commit_hook_path.write_text(
                f"""!#{POETRY_SHEBANG}

import sys

try:
    from autohooks.precommit import run
    sys.exit(run())
except ImportError:
    pass
            """,
                encoding="utf8",
            )
            pre_commit_hook = PreCommitHook()

            check_pre_commit_hook(term, pre_commit_hook)

        term.ok.assert_called_once_with("autohooks pre-commit hook is active.")
        term.warning.assert_called_once_with(
            "autohooks pre-commit hook is outdated. Please run "
            "'autohooks activate --force' to update your pre-commit "
            "hook."
        )
        term.info.assert_not_called()
        term.error.assert_not_called()

    def test_unknown_mode(self):
        term = MagicMock()

        with tempgitdir():
            pre_commit_hook_path = get_pre_commit_hook_path()
            pre_commit_hook_path.write_text(
                f"""#!/bin/sh
# meta = {{ version = {TEMPLATE_VERSION} }}

import sys

try:
    from autohooks.precommit import run
    sys.exit(run())
except ImportError:
    pass
            """,
                encoding="utf8",
            )
            pre_commit_hook = PreCommitHook()

            check_pre_commit_hook(term, pre_commit_hook)

        self.assertEqual(term.ok.call_count, 2)
        term.ok.assert_has_calls(
            (
                call("autohooks pre-commit hook is active."),
                call("autohooks pre-commit hook is up-to-date."),
            )
        )
        term.warning.assert_called_once_with(
            f"Unknown autohooks mode in {pre_commit_hook}. "
            f'Falling back to "pythonpath" mode.'
        )
        term.info.assert_not_called()
        term.error.assert_not_called()

    def test_is_active(self):
        term = MagicMock()

        with tempgitdir():
            pre_commit_hook = PreCommitHook()
            pre_commit_hook.write(mode=Mode.POETRY)

            check_pre_commit_hook(term, pre_commit_hook)

        self.assertEqual(term.ok.call_count, 2)
        term.ok.assert_has_calls(
            (
                call("autohooks pre-commit hook is active."),
                call("autohooks pre-commit hook is up-to-date."),
            )
        )
        term.warning.assert_not_called()
        term.info.assert_not_called()
        term.error.assert_not_called()


class CheckConfigTestCase(unittest.TestCase):
    def test_no_pyproject_toml(self):
        term = MagicMock()

        with tempgitdir():
            pre_commit_hook = PreCommitHook()
            pyproject_toml = get_pyproject_toml_path()

            check_config(term, pyproject_toml, pre_commit_hook)

        term.ok.assert_not_called()
        term.warning.assert_not_called()
        term.info.assert_not_called()
        term.error.assert_called_once_with(
            f"Missing {pyproject_toml} file. Please add a "
            'pyproject.toml file and include a "tool.autohooks" '
            "section."
        )

    def test_no_autohooks_section(self):
        term = MagicMock()

        with tempgitdir():
            pre_commit_hook = PreCommitHook()
            pyproject_toml = get_pyproject_toml_path()
            pyproject_toml.touch()

            check_config(term, pyproject_toml, pre_commit_hook)

        term.ok.assert_not_called()
        term.warning.assert_not_called()
        term.info.assert_not_called()
        term.error.assert_called_once_with(
            f"autohooks is not enabled in your {pyproject_toml} file."
            f' Please add a "{AUTOHOOKS_SECTION}" section.'
        )

    def test_undefined_mode(self):
        term = MagicMock()

        with tempgitdir():
            pre_commit_hook = PreCommitHook()
            pre_commit_hook.write(mode=Mode.POETRY)
            pyproject_toml = get_pyproject_toml_path()
            pyproject_toml.write_text(
                "[tool.autohooks]\npre-commit = []", encoding="utf8"
            )

            check_config(term, pyproject_toml, pre_commit_hook)

        term.ok.assert_not_called()
        term.warning.assert_called_once_with(
            f"autohooks mode is not defined in {pyproject_toml}."
        )
        term.info.assert_called_once_with('Using autohooks mode "poetry".')
        term.error.assert_called_once_with(
            "No autohooks plugin is activated in "
            f"{pyproject_toml} for your pre commit hook. Please "
            'add a "pre-commit = [plugin1, plugin2]" setting.'
        )

    def test_unknown_mode(self):
        term = MagicMock()

        with tempgitdir():
            pre_commit_hook = PreCommitHook()
            pre_commit_hook.write(mode=Mode.POETRY)
            pyproject_toml = get_pyproject_toml_path()
            pyproject_toml.write_text(
                "[tool.autohooks]\nmode = 'foo'\npre-commit = []",
                encoding="utf8",
            )

            check_config(term, pyproject_toml, pre_commit_hook)

        term.ok.assert_not_called()
        term.warning.assert_called_once_with(
            f"Unknown autohooks mode in {pyproject_toml}."
        )
        term.info.assert_called_once_with('Using autohooks mode "poetry".')
        term.error.assert_called_once_with(
            "No autohooks plugin is activated in "
            f"{pyproject_toml} for your pre commit hook. Please "
            'add a "pre-commit = [plugin1, plugin2]" setting.'
        )

    def test_different_mode(self):
        term = MagicMock()

        with tempgitdir():
            pre_commit_hook = PreCommitHook()
            pre_commit_hook.write(mode=Mode.POETRY)
            pyproject_toml = get_pyproject_toml_path()
            pyproject_toml.write_text(
                "[tool.autohooks]\nmode = 'pythonpath'\npre-commit = []",
                encoding="utf8",
            )

            check_config(term, pyproject_toml, pre_commit_hook)

        term.ok.assert_not_called()
        term.warning.assert_called_once_with(
            f'autohooks mode "poetry" in pre-commit '
            f"hook {pre_commit_hook} differs from "
            f'mode "pythonpath" in {pyproject_toml}.'
        )
        term.info.assert_called_once_with('Using autohooks mode "poetry".')
        term.error.assert_called_once_with(
            "No autohooks plugin is activated in "
            f"{pyproject_toml} for your pre commit hook. Please "
            'add a "pre-commit = [plugin1, plugin2]" setting.'
        )

    def test_no_plugin(self):
        term = MagicMock()

        with tempgitdir():
            pre_commit_hook = PreCommitHook()
            pre_commit_hook.write(mode=Mode.POETRY)
            pyproject_toml = get_pyproject_toml_path()
            pyproject_toml.write_text(
                "[tool.autohooks]\nmode = 'poetry'\npre-commit = []",
                encoding="utf8",
            )

            check_config(term, pyproject_toml, pre_commit_hook)

        term.ok.assert_not_called()
        term.warning.assert_not_called()
        term.info.assert_called_once_with('Using autohooks mode "poetry".')
        term.error.assert_called_once_with(
            "No autohooks plugin is activated in "
            f"{pyproject_toml} for your pre commit hook. Please "
            'add a "pre-commit = [plugin1, plugin2]" setting.'
        )

    def test_plugin_not_loadable(self):
        term = MagicMock()

        with tempgitdir():
            pre_commit_hook = PreCommitHook()
            pre_commit_hook.write(mode=Mode.POETRY)
            pyproject_toml = get_pyproject_toml_path()
            pyproject_toml.write_text(
                "[tool.autohooks]\nmode = 'poetry'\npre-commit = ['plugin1']",
                encoding="utf8",
            )

            check_config(term, pyproject_toml, pre_commit_hook)

        term.ok.assert_not_called()
        term.warning.assert_not_called()
        term.info.assert_called_once_with('Using autohooks mode "poetry".')
        term.error.assert_called_once_with(
            '"plugin1" is not a valid autohooks plugin. '
            "No module named 'plugin1'"
        )

    def test_plugin_no_precommit_function(self):
        term = MagicMock()

        with tempgitdir() as tmpdir:
            pre_commit_hook = PreCommitHook()
            pre_commit_hook.write(mode=Mode.POETRY)
            pyproject_toml = get_pyproject_toml_path()
            pyproject_toml.write_text(
                "[tool.autohooks]\nmode = 'poetry'\npre-commit = ['plugin1']",
                encoding="utf8",
            )
            dot_autohooks_dir = tmpdir / ".autohooks"
            dot_autohooks_dir.mkdir()
            plugin1 = dot_autohooks_dir / "plugin1.py"
            plugin1.touch()

            check_config(term, pyproject_toml, pre_commit_hook)

        term.ok.assert_not_called()
        term.warning.assert_not_called()
        term.info.assert_called_once_with('Using autohooks mode "poetry".')
        term.error.assert_called_once_with(
            'Plugin "plugin1" has no precommit '
            "function. The function is required to run"
            " the plugin as git pre commit hook."
        )

        del sys.modules["plugin1"]

    def test_plugin_old_precommit_signature(self):
        term = MagicMock()

        with tempgitdir() as tmpdir:
            pre_commit_hook = PreCommitHook()
            pre_commit_hook.write(mode=Mode.POETRY)
            pyproject_toml = get_pyproject_toml_path()
            pyproject_toml.write_text(
                "[tool.autohooks]\nmode = 'poetry'\npre-commit = ['plugin1']",
                encoding="utf8",
            )
            dot_autohooks_dir = tmpdir / ".autohooks"
            dot_autohooks_dir.mkdir()
            plugin1 = dot_autohooks_dir / "plugin1.py"
            plugin1.write_text(
                """
def precommit():
    pass
            """,
                encoding="utf8",
            )

            check_config(term, pyproject_toml, pre_commit_hook)

        term.ok.assert_not_called()
        term.warning.assert_called_once_with(
            'Plugin "plugin1" uses a deprecated '
            "signature for its precommit function. It "
            "is missing the **kwargs parameter."
        )
        term.info.assert_called_once_with('Using autohooks mode "poetry".')
        term.error.assert_not_called()

        del sys.modules["plugin1"]

    def test_success(self):
        term = MagicMock()

        with tempgitdir() as tmpdir:
            pre_commit_hook = PreCommitHook()
            pre_commit_hook.write(mode=Mode.POETRY)
            pyproject_toml = get_pyproject_toml_path()
            pyproject_toml.write_text(
                "[tool.autohooks]\nmode = 'poetry'\npre-commit = ['plugin1']",
                encoding="utf8",
            )
            dot_autohooks_dir = tmpdir / ".autohooks"
            dot_autohooks_dir.mkdir()
            plugin1 = dot_autohooks_dir / "plugin1.py"
            plugin1.write_text(
                """
def precommit(*args):
    pass
            """,
                encoding="utf8",
            )

            check_config(term, pyproject_toml, pre_commit_hook)

        term.ok.assert_called_once_with('Plugin "plugin1" active and loadable.')
        term.warning.assert_not_called()
        term.info.assert_called_once_with('Using autohooks mode "poetry".')
        term.error.assert_not_called()

        del sys.modules["plugin1"]



# SPDX-FileCopyrightText: 2019-2024 Greenbone AG
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

import re
from pathlib import Path
from typing import Optional

from autohooks.settings import Mode
from autohooks.template import (
    PIPENV_MULTILINE_SHEBANG,
    PIPENV_SHEBANG,
    POETRY_MULTILINE_SHEBANG,
    POETRY_SHEBANG,
    PYTHON3_SHEBANG,
    TEMPLATE_VERSION,
    PreCommitTemplate,
)
from autohooks.utils import get_git_hook_directory_path


def get_pre_commit_hook_path():
    git_hook_dir_path = get_git_hook_directory_path()
    return git_hook_dir_path / "pre-commit"


class PreCommitHook:
    def __init__(self, pre_commit_hook_path: Optional[Path] = None) -> None:
        self._pre_commit_hook = None

        if pre_commit_hook_path is None:
            self.pre_commit_hook_path = get_pre_commit_hook_path()
        else:
            self.pre_commit_hook_path = pre_commit_hook_path

    @property
    def pre_commit_hook(self) -> str:
        if self._pre_commit_hook is None:
            self._pre_commit_hook = self.pre_commit_hook_path.read_text()

        return self._pre_commit_hook  # type: ignore

    def exists(self) -> bool:
        return self.pre_commit_hook_path.exists()

    def is_autohooks_pre_commit_hook(self) -> bool:
        lines = self.pre_commit_hook.split("\n")
        # seems to be false-positive ...
        return len(lines) > 5 and "autohooks.precommit" in self.pre_commit_hook

    def is_current_autohooks_pre_commit_hook(self) -> bool:
        return self.read_version() == TEMPLATE_VERSION

    def read_mode(self) -> Mode:
        lines = self.pre_commit_hook.split("\n")
        if len(lines) < 1 or len(lines[0]) == 0:
            return Mode.UNDEFINED

        shebang = lines[0][2:]

        if shebang == PYTHON3_SHEBANG:
            return Mode.PYTHONPATH
        if shebang.startswith(POETRY_SHEBANG):
            return Mode.POETRY
        if shebang == PIPENV_SHEBANG:
            return Mode.PIPENV

        shebang = f"{lines[0][2:]}\n"
        shebang += "\n".join(lines[1:5])
        if shebang == POETRY_MULTILINE_SHEBANG:
            return Mode.POETRY_MULTILINE

        if shebang == PIPENV_MULTILINE_SHEBANG:
            return Mode.PIPENV_MULTILINE

        return Mode.UNKNOWN

    def read_version(self) -> int:
        matches = re.search(
            r"{\s*version\s*=\s*?(\d+)\s*}$", self.pre_commit_hook, re.MULTILINE
        )
        if not matches:
            return -1

        return int(matches.group(1))

    def write(self, *, mode: Mode) -> None:
        template = PreCommitTemplate()
        pre_commit_hook = template.render(mode=mode)

        self.pre_commit_hook_path.write_text(pre_commit_hook)
        self.pre_commit_hook_path.chmod(0o775)

        self._pre_commit_hook = None

    def __str__(self) -> str:
        return str(self.pre_commit_hook_path)



# SPDX-FileCopyrightText: 2019-2024 Greenbone AG
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

import shlex
import subprocess
from pathlib import Path
from typing import List, Optional


class GitError(subprocess.CalledProcessError):
    """
    Error raised if a git command fails.
    """

    def __str__(self) -> str:
        return (
            f"Git command '{self.cmd}' returned "
            f"non-zero exit status {str(self.returncode)}"
        )


def exec_git(*args: str, ignore_errors: bool = False) -> str:
    """
    Execute git command.

    Raises:
        GitError: A GitError is raised if the git commit fails and ignore_errors
            is False.

    Args:
        *args: Variable length argument list passed to git.
        ignore_errors: Ignore errors if git command fails. Default: False.

    Example: ::

        exec_git("commit", "-m", "A new commit")
    """
    try:
        cmd_args: List[str] = ["git"]
        cmd_args.extend(args)
        process = subprocess.run(
            cmd_args, check=True, capture_output=True, text=True
        )
        return process.stdout
    except subprocess.CalledProcessError as e:
        if ignore_errors:
            return ""
        raise GitError(e.returncode, e.cmd, e.output, e.stderr) from None


def get_git_directory_path() -> Path:
    pwd = Path.cwd()

    git_dir = exec_git("-C", str(pwd), "rev-parse", "--git-dir").rstrip()

    if pwd and str(pwd) not in git_dir:
        git_dir_path = Path(pwd) / git_dir
    else:
        git_dir_path = Path(git_dir)

    return git_dir_path.resolve()


def get_autohooks_directory_path() -> Path:
    return Path(__file__).resolve().parent


def get_git_hook_directory_path(git_dir_path: Optional[Path] = None) -> Path:
    if git_dir_path is None:
        git_dir_path = get_git_directory_path()
    return git_dir_path / "hooks"


def is_project_root(path: Path) -> bool:
    return (
        (path / "pyproject.toml").is_file()
        or (path / ".git").is_dir()
        or (path / "setup.py").is_file()
        or (path / "setup.cfg").is_file()
    )


def get_project_root_path(path: Optional[Path] = None) -> Path:
    if path is None:
        path = Path.cwd()

    path.resolve()

    if is_project_root(path):
        return path

    for parent in path.parents:
        if is_project_root(parent):
            return parent

    return path


def get_project_autohooks_plugins_path(path: Optional[Path] = None) -> Path:
    root = get_project_root_path(path)
    return root / ".autohooks"


def get_pyproject_toml_path(path: Optional[Path] = None) -> Path:
    root = get_project_root_path(path)
    return root / "pyproject.toml"


def is_split_env():
    try:
        subprocess.run(
            shlex.split("/usr/bin/env -S echo True"),
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            universal_newlines=True,
            check=True,
        )
        is_split = True
    except (subprocess.CalledProcessError, FileNotFoundError):
        is_split = False

    return is_split



# SPDX-FileCopyrightText: 2019-2024 Greenbone AG
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

import argparse
import sys

import shtab

from autohooks.__version__ import __version__ as version
from autohooks.cli.activate import install_hooks
from autohooks.cli.check import check_hooks
from autohooks.cli.plugins import (
    add_plugins,
    list_plugins,
    plugins,
    remove_plugins,
)
from autohooks.settings import Mode
from autohooks.terminal import Terminal

DESCRIPTION = "autohooks - Manage git hooks"


def main():
    parser = argparse.ArgumentParser(description=DESCRIPTION)
    shtab.add_argument_to(parser)
    parser.add_argument(
        "--version",
        action="version",
        version=f"%(prog)s {version}",
    )

    subparsers = parser.add_subparsers(dest="command")
    activate_parser = subparsers.add_parser(
        "activate", help="Activate the pre-commit hook."
    )
    activate_parser.add_argument(
        "-f",
        "--force",
        action="store_true",
        help="Force activation of hook even if a hook already exists",
    )
    activate_parser.add_argument(
        "-m",
        "--mode",
        dest="mode",
        choices=[
            str(Mode.PYTHONPATH),
            str(Mode.PIPENV),
            str(Mode.POETRY),
        ],
        help="Mode for loading autohooks during hook execution. Either load "
        "autohooks from the PYTHON_PATH, via pipenv or via poetry.",
    )
    activate_parser.set_defaults(func=install_hooks)

    check_parser = subparsers.add_parser(
        "check", help="Check installed pre-commit hook"
    )
    check_parser.set_defaults(func=check_hooks)

    plugins_parser = subparsers.add_parser(
        "plugins", help="Manage autohooks plugins"
    )
    plugins_parser.set_defaults(func=plugins)

    plugins_subparsers = plugins_parser.add_subparsers(
        dest="subcommand", required=True
    )

    add_plugins_parser = plugins_subparsers.add_parser(
        "add", help="Add plugins."
    )
    add_plugins_parser.set_defaults(plugins_func=add_plugins)
    add_plugins_parser.add_argument("name", nargs="+", help="Plugin(s) to add")

    remove_plugins_parser = plugins_subparsers.add_parser(
        "remove", help="Remove plugins."
    )
    remove_plugins_parser.set_defaults(plugins_func=remove_plugins)
    remove_plugins_parser.add_argument(
        "name", nargs="+", help="Plugin(s) to remove"
    )

    list_plugins_parser = plugins_subparsers.add_parser(
        "list", help="List current used plugins."
    )
    list_plugins_parser.set_defaults(plugins_func=list_plugins)

    args = parser.parse_args()

    if not args.command:
        parser.print_usage()
        sys.exit(1)

    term = Terminal()
    args.func(term, args)


if __name__ == "__main__":
    main()



# SPDX-FileCopyrightText: 2022-2024 Greenbone AG
#
# SPDX-License-Identifier: GPL-3.0-or-later
#


import unittest

from autohooks.precommit.run import (
    CheckPluginError,
    CheckPluginWarning,
    check_plugin,
)
from tests import temp_python_module, tempdir


class CheckPluginTestCase(unittest.TestCase):
    def test_plugin_not_found(self):
        with tempdir(change_into=True):
            result = check_plugin("foo.bar")

            self.assertIsInstance(result, CheckPluginError)
            self.assertIn(
                '"foo.bar" is not a valid autohooks plugin.', result.message
            )

    def test_no_precommit_function(self):
        content = """print()"""
        with temp_python_module(
            content,
            name="foo",
        ):
            result = check_plugin("foo")

            self.assertIsInstance(result, CheckPluginError)
            self.assertEqual(
                result.message,
                'Plugin "foo" has no precommit function. The function is '
                "required to run the plugin as git pre commit hook.",
            )

    def test_no_precommit_function_arguments(self):
        content = """def precommit():
  print()"""
        with temp_python_module(
            content,
            name="foo",
        ):
            result = check_plugin("foo")

            self.assertIsInstance(result, CheckPluginWarning)
            self.assertEqual(
                'Plugin "foo" uses a deprecated signature for its precommit '
                "function. It is missing the **kwargs parameter.",
                result.message,
            )

    def test_success(self):
        content = """def precommit(**kwargs):
  print()"""
        with temp_python_module(
            content,
            name="foo",
        ):
            self.assertIsNone(check_plugin("foo"))



# SPDX-FileCopyrightText: 2019-2024 Greenbone AG
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

import unittest

from autohooks.settings import Mode
from autohooks.template import (
    PreCommitTemplate,
    get_pre_commit_hook_template_path,
)

DEFAULT_TEMPLATE = """#!/usr/bin/env python3
# meta = { version = 1 }

import sys

try:
    from autohooks.precommit import run
    sys.exit(run())
except ImportError:
    print(
        "Error: autohooks is not installed. To force creating a commit without "
        "verification via autohooks run 'git commit --no-verify'.",
        file=sys.stderr,
    )
    sys.exit(1)
"""


class FakeTemplatePath:
    def __init__(self, text):
        self._text = text

    def read_text(self):
        return self._text


class PreCommitTemplateTestCase(unittest.TestCase):
    def test_should_use_default_template(self):
        template = PreCommitTemplate()
        self.assertEqual(
            template.render(mode=Mode.PYTHONPATH), DEFAULT_TEMPLATE
        )

    def test_should_render_mode_pipenv(self):
        path = FakeTemplatePath("$SHEBANG")
        template = PreCommitTemplate(path)
        self.assertEqual(
            template.render(mode=Mode.PIPENV),
            "/usr/bin/env -S pipenv run python3",
        )

    def test_should_render_mode_poetry(self):
        path = FakeTemplatePath("$SHEBANG")
        template = PreCommitTemplate(path)
        self.assertEqual(
            template.render(mode=Mode.POETRY),
            "/usr/bin/env -S poetry run python",
        )

    def test_should_render_mode_pipenv_multiline(self):
        path = FakeTemplatePath("$SHEBANG")
        template = PreCommitTemplate(path)
        self.assertEqual(
            template.render(mode=Mode.PIPENV_MULTILINE),
            (
                "/bin/sh\n"
                "\"true\" ''':'\n"
                'pipenv run python3 "$0" "$@"\n'
                'exit "$?"\n'
                "'''"
            ),
        )

    def test_should_render_mode_poetry_multiline(self):
        path = FakeTemplatePath("$SHEBANG")
        template = PreCommitTemplate(path)
        self.assertEqual(
            template.render(mode=Mode.POETRY_MULTILINE),
            (
                "/bin/sh\n"
                "\"true\" ''':'\n"
                'poetry run python "$0" "$@"\n'
                'exit "$?"\n'
                "'''"
            ),
        )

    def test_should_render_mode_unknown(self):
        path = FakeTemplatePath("$SHEBANG")
        template = PreCommitTemplate(path)
        self.assertEqual(
            template.render(mode=Mode.UNDEFINED), "/usr/bin/env python3"
        )

    def test_should_render_mode_undefined(self):
        path = FakeTemplatePath("$SHEBANG")
        template = PreCommitTemplate(path)
        self.assertEqual(
            template.render(mode=Mode.UNDEFINED), "/usr/bin/env python3"
        )


class GetPreCommitHookTemplatePath(unittest.TestCase):
    def test_template_exists(self):
        template_path = get_pre_commit_hook_template_path()
        self.assertTrue(template_path.exists())
        self.assertTrue(template_path.is_file())


if __name__ == "__main__":
    unittest.main()



# SPDX-FileCopyrightText: 2019-2024 Greenbone AG
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

from pathlib import Path
from string import Template
from typing import Dict, Optional, Union

from autohooks.settings import Mode
from autohooks.utils import get_autohooks_directory_path

PYTHON3_SHEBANG = "/usr/bin/env python3"
PIPENV_SHEBANG = "/usr/bin/env -S pipenv run python3"
POETRY_SHEBANG = "/usr/bin/env -S poetry run python"
# For OS's that don't support '/usr/bin/env -S'.
PIPENV_MULTILINE_SHEBANG = (
    "/bin/sh\n"
    "\"true\" ''':'\n"
    'pipenv run python3 "$0" "$@"\n'
    'exit "$?"\n'
    "'''"
)
POETRY_MULTILINE_SHEBANG = (
    "/bin/sh\n"
    "\"true\" ''':'\n"
    'poetry run python "$0" "$@"\n'
    'exit "$?"\n'
    "'''"
)

TEMPLATE_VERSION = 1


def get_pre_commit_hook_template_path() -> Path:
    setup_dir_path = get_autohooks_directory_path()
    return setup_dir_path / "precommit" / "template"


class PreCommitTemplate:
    def __init__(self, template_path: Optional[Path] = None) -> None:
        if template_path is None:
            template_path = get_pre_commit_hook_template_path()
        self._load(template_path)

    def _load(self, template_path: Path) -> None:
        self._template = Template(template_path.read_text())

    def render(self, *, mode: Mode) -> str:
        mode = mode.get_effective_mode()

        params: Dict[str, Union[str, int]] = dict(VERSION=TEMPLATE_VERSION)

        if mode == Mode.PIPENV:
            params["SHEBANG"] = PIPENV_SHEBANG
        elif mode == Mode.POETRY:
            params["SHEBANG"] = POETRY_SHEBANG
        elif mode == Mode.PIPENV_MULTILINE:
            params["SHEBANG"] = PIPENV_MULTILINE_SHEBANG
        elif mode == Mode.POETRY_MULTILINE:
            params["SHEBANG"] = POETRY_MULTILINE_SHEBANG
        else:
            params["SHEBANG"] = PYTHON3_SHEBANG

        return self._template.safe_substitute(params)



# SPDX-FileCopyrightText: 2022-2024 Greenbone AG
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

from autohooks.api.git import Status, get_status, stash_unstaged_changes
from tests import tempgitdir

from . import GitTestCase, git_add, randbytes


class StashUnstagedChangesTestCase(GitTestCase):
    def test_no_working_files(self):
        with tempgitdir() as tmpdir:
            file1 = tmpdir / "foo.txt"
            file1.write_bytes(randbytes(20))

            git_add(file1)

            stash = stash_unstaged_changes()
            with stash:
                self.assertEqual(len(stash.partially_staged), 0)

    def test_staged_with_working_files(self):
        with tempgitdir() as tmpdir:
            file1 = tmpdir / "foo.txt"
            file1.write_bytes(randbytes(20))

            git_add(file1)

            file1 = tmpdir / "bar.txt"
            file1.write_bytes(randbytes(20))

            stash = stash_unstaged_changes()
            with stash:
                self.assertEqual(len(stash.partially_staged), 0)

    def test_partiall_staged_files(self):
        with tempgitdir() as tmpdir:
            file1 = tmpdir / "foo.txt"
            file1.write_bytes(randbytes(20))

            git_add(file1)

            file1.write_bytes(randbytes(20))

            status = get_status()
            self.assertEqual(status[0].index, Status.ADDED)
            self.assertEqual(status[0].working_tree, Status.MODIFIED)

            stash = stash_unstaged_changes()
            with stash:
                self.assertEqual(len(stash.partially_staged), 1)

                status = get_status()
                self.assertEqual(status[0].index, Status.ADDED)
                self.assertEqual(status[0].working_tree, Status.UNMODIFIED)

            status = get_status()
            self.assertEqual(status[0].index, Status.ADDED)
            self.assertEqual(status[0].working_tree, Status.MODIFIED)

    def test_partiall_staged_files_with_error(self):
        with tempgitdir() as tmpdir:
            file1 = tmpdir / "foo.txt"
            file1.write_bytes(randbytes(20))

            git_add(file1)

            content = randbytes(20)
            file1.write_bytes(content)

            status = get_status()
            self.assertEqual(status[0].index, Status.ADDED)
            self.assertEqual(status[0].working_tree, Status.MODIFIED)

            stash = stash_unstaged_changes()

            with self.assertRaises(ValueError):
                with stash:
                    self.assertEqual(len(stash.partially_staged), 1)

                    status = get_status()
                    self.assertEqual(status[0].index, Status.ADDED)
                    self.assertEqual(status[0].working_tree, Status.UNMODIFIED)

                    raise ValueError("An error ocurred!")

            status = get_status()
            self.assertEqual(status[0].index, Status.ADDED)
            self.assertEqual(status[0].working_tree, Status.MODIFIED)

            self.assertEqual(content, file1.read_bytes())

    def test_partiall_staged_files_with_error_and_changed_content(self):
        with tempgitdir() as tmpdir:
            file1 = tmpdir / "foo.txt"
            file1.write_bytes(randbytes(20))

            git_add(file1)

            content = randbytes(20)
            file1.write_bytes(content)

            status = get_status()
            self.assertEqual(status[0].index, Status.ADDED)
            self.assertEqual(status[0].working_tree, Status.MODIFIED)

            stash = stash_unstaged_changes()

            with self.assertRaises(ValueError):
                with stash:
                    self.assertEqual(len(stash.partially_staged), 1)

                    status = get_status()
                    self.assertEqual(status[0].index, Status.ADDED)
                    self.assertEqual(status[0].working_tree, Status.UNMODIFIED)

                    content2 = randbytes(20)
                    file1.write_bytes(content2)

                    raise ValueError("An error ocurred!")

            status = get_status()
            self.assertEqual(status[0].index, Status.ADDED)
            self.assertEqual(status[0].working_tree, Status.MODIFIED)

            self.assertEqual(content, file1.read_bytes())

    def test_formatting_plugin_with_untracked_change(self):
        with tempgitdir() as tmpdir:
            content = "Lorem Ipsum"
            content2 = "Lorem Ipsum\nDolor Sit"
            content3 = "Lorem Ipsum\nDolor Sit\namet consetetur sadipscing"

            file1 = tmpdir / "foo.txt"
            file1.write_text(content, encoding="utf8")

            git_add(file1)

            file1.write_text(content2, encoding="utf8")

            stash = stash_unstaged_changes()
            with stash:
                self.assertEqual(len(stash.partially_staged), 1)
                self.assertEqual(content, file1.read_text(encoding="utf8"))

                file1.write_text(content3, encoding="utf8")

            self.assertEqual(content2, file1.read_text(encoding="utf8"))

    def test_formatting_plugin_with_staged_change(self):
        with tempgitdir() as tmpdir:
            content = "Lorem Ipsum"
            content2 = "Lorem Ipsum\nDolor Sit"

            file1 = tmpdir / "foo.txt"
            file1.write_text(content, encoding="utf8")

            git_add(file1)

            file1.write_text(content2, encoding="utf8")

            stash = stash_unstaged_changes()
            with stash:
                self.assertEqual(len(stash.partially_staged), 1)
                self.assertEqual(content, file1.read_text(encoding="utf8"))

                file1.write_text(content2, encoding="utf8")

                git_add(file1)

            self.assertEqual(content2, file1.read_text(encoding="utf8"))

    def test_formatting_plugin_with_conflicting_staged_change(self):
        with tempgitdir() as tmpdir:
            content = "Lorem Ipsum\n\n\n"
            content2 = "Lorem Ipsum\nDolor Sit\n\n"
            content3 = "Lorem Ipsum\n\nAmet\n"
            content4 = "Lorem Ipsum\nDolor Sit\nAmet\n"

            file1 = tmpdir / "foo.txt"
            file1.write_text(content, encoding="utf8")

            git_add(file1)

            file1.write_text(content2, encoding="utf8")

            stash = stash_unstaged_changes()
            with stash:
                self.assertEqual(len(stash.partially_staged), 1)
                self.assertEqual(content, file1.read_text(encoding="utf8"))

                file1.write_text(content3, encoding="utf8")

                git_add(file1)

            self.assertEqual(content4, file1.read_text(encoding="utf8"))



# SPDX-FileCopyrightText: 2022-2024 Greenbone AG
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

import random
import unittest
from pathlib import Path

from autohooks.utils import exec_git


def randbytes(n: int) -> bytes:  # pylint: disable=invalid-name
    if hasattr(random, "randbytes"):
        return random.randbytes(n)
    return random.getrandbits(n * 8).to_bytes(n, "little")


def git_add(*paths: Path) -> None:
    exec_git("add", *paths)  # type: ignore[arg-type]


def git_rm(*paths: Path) -> None:
    exec_git("rm", *paths)  # type: ignore[arg-type]


def git_mv(from_path: Path, to_path: Path) -> None:
    exec_git("mv", from_path, to_path)  # type: ignore[arg-type]


def git_commit(message: str = "Foo Bar"):
    exec_git("commit", "--no-gpg-sign", "-m", message)


class GitTestCase(unittest.TestCase):
    pass



# SPDX-FileCopyrightText: 2019-2024 Greenbone AG
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

import unittest
from unittest.mock import Mock

from autohooks.api import error, fail, info, ok, out, warning
from autohooks.terminal import Terminal, _set_terminal


class TerminalOutputApiTestCase(unittest.TestCase):
    def setUp(self):
        self.term = Mock(spec=Terminal)
        _set_terminal(self.term)

    def test_error(self):
        error("foo bar")
        self.term.error.assert_called_with("foo bar")

    def test_fail(self):
        fail("foo bar")
        self.term.fail.assert_called_with("foo bar")

    def test_info(self):
        info("foo bar")
        self.term.info.assert_called_with("foo bar")

    def test_ok(self):
        ok("foo bar")
        self.term.ok.assert_called_with("foo bar")

    def test_out(self):
        out("foo bar")
        self.term.out.assert_called_with("foo bar")

    def test_warning(self):
        warning("foo bar")
        self.term.warning.assert_called_with("foo bar")


if __name__ == "__main__":
    unittest.main()
